"""Implements a client for making a udp connection to the duka one devices """
import socket
import threading
import time

from socket import SOL_SOCKET, SO_REUSEADDR, SO_BROADCAST

from .device import Device, Mode, Speed
from .dukapacket import DukaPacket

class DukaClient:
    """Client object for making connection to the duka devices."""
    _mutex = threading.Lock()

    def __init__(self):
        self._devices = {}
        self._sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self._sock.setsockopt(SOL_SOCKET, SO_REUSEADDR, 1)
        self._sock.setsockopt(SOL_SOCKET, SO_BROADCAST, 1)

        self._notifyrunning = False
        self._notifythread = threading.Thread(target=self.__notify_fn)
        self._notifythread.start()

    def close(self):
        """Close the client and end the notify thread. Wait for the thread to end."""
        self._notifyrunning = False
        self._notifythread.join()

    def add_device(self, device_id: str, password: str = None,
                   ip_address: str = "<broadcast>", onchange=None) -> Device:
        """Add a new device. If the device already exist the current one will
        be returned"""
        device: Device = self.get_device(device_id)
        if device is None:
            device = Device(device_id, password, ip_address, onchange)
            self._devices[device_id] = device
        return device

    def remove_device(self, device_id):
        """Remove an existing device"""
        device: Device = self.get_device(device_id)
        if device is not None:
            del self._devices[device_id]
        return device

    def get_device(self, device_id: str) -> Device:
        """Get a device by device id."""
        if not device_id in self._devices:
            return None
        return self._devices[device_id]

    def get_device_count(self):
        """Return the number of devices"""
        return len(self._devices)

    def set_speed(self, device: Device, speed: Speed):
        """Set the speed of the specified device"""
        if device.speed == speed:
            return
        if speed == Speed.OFF:
            self.turn_off(device)
            return
        if device.speed == Speed.OFF:
            self.turn_on(device)
            time.sleep(0.2)

        packet = DukaPacket()
        packet.initialize_speed_cmd(device, speed)
        data = packet.data
        self.__send_data(device, data)

    def turn_off(self, device: Device):
        """Turn off the specified device"""
        packet = DukaPacket()
        packet.initialize_off_cmd(device)
        data = packet.data
        self.__send_data(device, data)


    def turn_on(self, device: Device):
        """Turn on the specified device"""
        packet = DukaPacket()
        packet.initialize_on_cmd(device)
        data = packet.data
        self.__send_data(device, data)

    def set_mode(self, device: Device, mode: Mode):
        """Set the mode of the specified device"""
        if device.mode == Mode:
            return
        packet = DukaPacket()
        packet.initialize_mode_cmd(device, mode)
        data = packet.data
        self.__send_data(device, data)

    def validate_device(self, device_id: str, password: str = None,
                        ip_address: str = "<broadcast>") -> Device:
        """Validate if a device exist and repsonds.
           Returns None if the device does not exist
           Returns the Device object if it exist
         """                    
        device: Device = self.get_device( device_id)                        
        # Is the device already added
        if device is not None:
            return device
        device = self.add_device(device_id,password,ip_address)            
        try:
            self.__update_device_status(device)
            # 2 sec timeout
            timeout = time.time() + 2
            while True:
                if device.mode is not None:
                    return device
                if time.time() > timeout:
                    break
                time.sleep( 0.1)
            return None                    
        finally:
            self.remove_device(device.device_id)

    def __update_device_status(self, device: Device):
        """ Update the device status from the DukaClient
            You should not call this youself
        """
        packet = DukaPacket()
        packet.initialize_status_cmd(device)
        data = packet.data
        self.__send_data(device, data)

    def __update_all_device_status(self):
        for device_id in self._devices:
            self.__update_device_status(self._devices[device_id])

    def __send_data(self, device: Device, data):
        with DukaClient._mutex:
            self._sock.sendto(data, (device.ip_address, 4000))

    def __print_data(self, data):
        """Print data in hex - for debugging purpose """
        print(''.join('{:02x}'.format(x) for x in data))


    def __notify_fn(self):
        self._notifyrunning = True
        self._sock.bind(("0.0.0.0", 4000))
        self._sock.settimeout(1.0)
        try:
            while self._notifyrunning:
                try:
                    data, addr = self._sock.recvfrom(1024)
                except socket.timeout:
                    self.__update_all_device_status()
                    continue
                packet = DukaPacket()
                if not packet.initialize_from_data(data):
                    continue
                if not packet.is_response_from_device():
                    continue
                device_id = packet.response_device_id()
                if device_id not in self._devices:
                    continue
                device: Device = self._devices[device_id]
                ip_address = addr[0]
                speed = packet.response_speed()
                if not packet.response_is_on():
                    speed = Speed.OFF
                mode = packet.response_mode()
                device.update(ip_address, speed, mode)
        finally:
            self._sock.close()



"""Implements a class for the UDP data packet"""
from  .device import Device, Mode, Speed

class DukaPacket:
    """A udp data packet to/from the duka device."""

    def __init__(self):
        self._data = None
        self._pos = 0
        self._datastart = 0

    def initialize_from_data(self, data) -> bool:
        """Initialize a packet from data revieved from the device"""
        self._data = data
        size = len(self._data)
        if size < 4:
            return False
        idlen = data[3]
        pswstart = 4 + idlen
        if size <= pswstart:
            return False
        pswlen = data[pswstart]
        self._datastart = pswstart+1+pswlen
        if size <= self._datastart:
            return False
        checksum = self.__calc_checksum()
        return self._data[size-2] == (checksum & 0xff) and self._data[size-1] == (checksum >> 8)


    def initialize_speed_cmd(self, device: Device, speed: Speed):
        """Initialize a speed command packet to be sent to a device"""
        self.__build_data(device.device_id, device.password, 8)
        self.__add_byte(0x03)
        self.__add_byte(0x02)
        self.__add_byte(speed)
        self.__add_byte(0xfc)
        self.__add_byte(0x01)
        self.__add_byte(0x07)
        self.__add_byte(0x0b)
        self.__add_byte(0x72)
        self.__add_checksum()


    def initialize_mode_cmd(self, device: Device, mode: Mode):
        """Intialize a mode command packet to be sent to a device"""
        self.__build_data(device.device_id, device.password, 3)
        self.__add_byte(0x03)
        self.__add_byte(0xb7)
        self.__add_byte(mode)
        self.__add_checksum()


    def initialize_on_cmd(self, device: Device):
        """Initialize a ON command packet to be sent to a device"""
        self.__build_data(device.device_id, device.password, 3)
        self.__add_byte(0x03)
        self.__add_byte(0x01)
        self.__add_byte(0x01)
        self.__add_checksum()


    def initialize_off_cmd(self, device: Device):
        """Initialize a Off command packet to be sent to a device"""
        self.__build_data(device.device_id, device.password, 10)
        self.__add_byte(0x03)
        self.__add_byte(0x01)
        self.__add_byte(0x00)
        self.__add_byte(0xfc)
        self.__add_byte(0x01)
        self.__add_byte(0x07)
        self.__add_byte(0x0b)
        self.__add_byte(0x06)
        self.__add_byte(0x83)
        self.__add_byte(0x72)
        self.__add_checksum()

    def initialize_status_cmd(self, device: Device):
        """Initialize a status command packet to be sent to a device"""
        self.__build_data(device.device_id, device.password, 17)
        self.__add_byte(0x01)
        self.__add_byte(0xb9)
        self.__add_byte(0x01)
        self.__add_byte(0x02)
        self.__add_byte(0x44)
        self.__add_byte(0xb7)
        self.__add_byte(0x72)
        self.__add_byte(0x07)
        self.__add_byte(0x0b)
        self.__add_byte(0x06)
        self.__add_byte(0x88)
        self.__add_byte(0x83)
        self.__add_byte(0x32)
        self.__add_byte(0xff)
        self.__add_byte(0x03)
        self.__add_byte(0x04)
        self.__add_byte(0x05)
        self.__add_checksum()

    @property
    def data(self):
        """Return the data for the packet"""
        return self._data

    def is_response_from_device(self) -> bool:
        """Return true of the packet is from a device"""
        responsetype = self._data[self._datastart]
        return  responsetype == 0x06

    def response_device_id(self) -> str:
        """Return the device id from the packet"""
        idlen = self._data[3]
        device_id = ""
        for i in range(4, 4+idlen):
            device_id += chr(self._data[i])
        return device_id

    def response_is_on(self) -> bool:
        """Return true if this is a response On packet"""
        on_packet = None
        if self._data[self._datastart+1] == 0xfe:
            on_packet = self._data[self._datastart + 7] == 0x01
        if self._data[self._datastart+1] == 0x02:
            return True
        return on_packet

    def response_speed(self) -> Speed:
        """Return the speed from the packet"""
        speed = None
        if self._data[self._datastart+1] == 0xfe:
            speed = self._data[self._datastart + 9]
        if self._data[self._datastart+1] == 0x02:
            speed = self._data[self._datastart + 2]
        return speed

    def response_mode(self) -> Mode:
        """Return the mode from the packet"""
        mode = None
        if self._data[self._datastart+1] == 0xfe:
            mode = self._data[self._datastart + 13]
        if self._data[self._datastart+1] == 0xb7:
            mode = self._data[self._datastart + 2]
        return mode

    def __add_byte(self, byte: int):
        """Add a byte to the packet"""
        self._data[self._pos] = byte
        self._pos += 1


    def __build_data(self, device_id: str, password: str, size: int):
        """Build a packet of the specified size"""
        totalsize = 5+len(device_id)+len(password)+size+2
        self._data = bytearray(totalsize)
        self._pos = 0
        self.__add_byte(0xfd)
        self.__add_byte(0xfd)
        self.__add_byte(0x02)
        self.__add_byte(len(device_id))
        for char in device_id:
            self.__add_byte(ord(char))
        self.__add_byte(len(password))
        for char in password:
            self.__add_byte(ord(char))
        self._datastart = self._pos



    def __add_checksum(self):
        """Add a checksum to the packet"""
        checksum = self. __calc_checksum()
        self.__add_byte(checksum & 0xff)
        self.__add_byte(checksum >> 8)


    def __calc_checksum(self) -> int:
        """Calculate the check sum for the packet"""
        checksum: int = 0
        for i in range(2, len(self._data)-2):
            checksum += self._data[i]
        return checksum & 0xffff



"""
Setup of the dukaonesdk module
"""
from setuptools import setup

setup(
    name='dukaonesdk',
    version='0.9.5',
    description='Duka One ventilation SDK',
    long_description=("SDK for connection to the Duka One S6W ventilation. "
                      "Made for interfacing to home assistant"),
    author='Jens Østergaard Nielsen',
    url='https://github.com/dingusdk/dukaonesdk',
    packages=['dukaonesdk'],
    license='GPL-3.0',
)



"""Implements the duka one device class """
from  enum import IntEnum

class Speed(IntEnum):
    """Device speed options available """
    OFF = 0
    LOW = 1
    MEDIUM = 2
    HIGH = 3


class Mode(IntEnum):
    """Device modes available
    Note: The ONEWAY derection is decided by the dip switch on the device
    """
    ONEWAY = 0
    TWOWAY = 1
    IN = 2


class Device:
    """A class representing a single Duke One Device """

    def __init__(self, deviceid: str, password: str = None,
                 ip_address: str = "<broadcast>", onchange=None):
        self._id = deviceid
        self._password = password
        self._ip = ip_address
        self._speed: Speed = None
        self._mode: Mode = None
        self._changeevent = onchange

    @property
    def device_id(self) -> str:
        """Return  the device id """
        return self._id

    @property
    def password(self) -> str:
        """Return the password for the device"""
        if self._password:
            return self._password
        return "1111"

    @property
    def ip_address(self) -> str:
        """Return the IP of the device"""
        return self._ip

    @property
    def speed(self) -> Speed:
        """Return the speed of the device"""
        return self._speed

    @property
    def mode(self)  -> Mode:
        """Return the mode of the device"""
        return self._mode

    def update(self, ip_address: str, speed: Speed, mode: Mode):
        """Update the device with data recieved. Called by the dukaclient"""
        haschange = False
        if ip_address is not None and ip_address != self._ip:
            self._ip = ip_address
            haschange = True
        if speed is not None and speed != self._speed:
            self._speed = speed
            haschange = True
        if mode is not None and mode != self._mode:
            self._mode = mode
            haschange = True
        if haschange and self._changeevent is not None:
            self._changeevent(self)
